#!/usr/bin/python

import os, sys, struct
import fft.map.resource

# I may figure out what these mean later, so I'm making constants for them:
INDEX1_22 = 0x22
INDEX1_30 = 0x30
INDEX1_70 = 0x70
ARRANGE_00 = 0x00
ARRANGE_01 = 0x01
ARRANGE_02 = 0x02
ARRANGE_03 = 0x03
ARRANGE_04 = 0x04
ARRANGE_05 = 0x05
WEATHER_DAY_00 = 0x00
WEATHER_DAY_10 = 0x10
WEATHER_DAY_20 = 0x20
WEATHER_DAY_30 = 0x30
WEATHER_DAY_40 = 0x40
WEATHER_NIGHT_80 = 0x80
WEATHER_NIGHT_90 = 0x90
WEATHER_NIGHT_A0 = 0xA0
WEATHER_NIGHT_B0 = 0xB0
WEATHER_NIGHT_C0 = 0xC0
DEFAULT_INDEX = (INDEX1_22, ARRANGE_00, WEATHER_DAY_00)

index1_names = {
    INDEX1_22: '22h',
    INDEX1_30: '30h',
    INDEX1_70: '70h',
}

arrangement_names = {
    ARRANGE_00: 'Arrangement 0',
    ARRANGE_01: 'Arrangement 1',
    ARRANGE_02: 'Arrangement 2',
    ARRANGE_03: 'Arrangement 3',
    ARRANGE_04: 'Arrangement 4',
    ARRANGE_05: 'Arrangement 5',
}

weather_names = {
    WEATHER_DAY_00: 'Day 00h',
    WEATHER_DAY_10: 'Day 10h',
    WEATHER_DAY_20: 'Day 20h',
    WEATHER_DAY_30: 'Day 30h',
    WEATHER_DAY_40: 'Day 40h',
    WEATHER_NIGHT_80: 'Night 80h',
    WEATHER_NIGHT_90: 'Night 90h',
    WEATHER_NIGHT_A0: 'Night A0h',
    WEATHER_NIGHT_B0: 'Night B0h',
    WEATHER_NIGHT_C0: 'Night C0h',
}

RESOURCE_EOF = 0x3101
RESOURCE_TEXTURE = 0x1701

class File(file):

    def __init__(self, gns_file_spec):
        self.texture_files = {}
        self.other_files = {}
        self.all_indexes = {}
        self.texture = None
        self.polygons = []
        self.polygons_instructions = []
        self.polygons_extra = []
        self.polygons_extra_transform = []
        self.lights = []
        self.background = []
        self.terrain = []
        self.palettes_color = []
        self.palettes_color_alpha = []

        """
        Find and read GNS file with the given filename. Gather information about
        the resources it contains.
        """
        if not os.path.isfile(gns_file_spec):
            gns_file_spec = gns_file_spec.lower()
        try:
            file.__init__(self, gns_file_spec, 'rb')
        except IOError:
            print 'Unable to open file', gns_file_spec
            sys.exit(1)
        map_path = os.path.dirname(gns_file_spec)
        (index1, index2, weather, resource_type) = \
                struct.unpack('<HBBH', self.read(6))
        while not resource_type == RESOURCE_EOF:
            self.seek(2, 1)
            resource_lba = struct.unpack('<I', self.read(4))[0]
            resource_file_name = fft.map.lbas[resource_lba]
            resource_file_spec = os.path.join(map_path, resource_file_name)
            self.all_indexes[(index1, index2, weather)] = True
            if resource_type == RESOURCE_TEXTURE:
                self.texture_files[(index1, index2, weather)] = resource_file_spec
            else:
                self.other_files[(index1, index2, weather)] = resource_file_spec
            self.seek(8, 1)
            (index1, index2, weather, resource_type) = \
                    struct.unpack('<HBBH', self.read(6))
        self.all_indexes = sorted(self.all_indexes.keys())

    def stringify_index(self, gns_index):
        index1, arrangement, weather = gns_index
        return "%s, %s, %s" % (
                index1_names[index1],
                arrangement_names[arrangement],
                weather_names[weather]
                )

    def get_resources(self, index):
        self.texture = self.get_texture(index)
        self.polygons = self.get_polygons(index, fft.map.TOC_MESH)
        i = 0
        for vis in self.get_polygon_visibility(index):
            self.polygons[i]['vis'] = vis
            i += 1
        for i in range(8):
            extras = self.get_polygons(index, fft.map.TOC_MESH_EXTRA0 + i * 4)
            if len(extras):
                self.polygons_extra.append(extras)
        self.get_polygons_anim(index)
        (self.palettes_color, self.palettes_color_alpha) = \
                self.get_palettes_color(index)
        self.lights = self.get_lights(index)
        self.background = self.get_background(index)
        self.terrain = self.get_terrain(index)

    def open_resource(self, index, resource_type):
        try_indexes = [
                index,
                (index[0], index[1], WEATHER_DAY_00),
                (index[0], ARRANGE_00, WEATHER_DAY_00),
                (INDEX1_30, ARRANGE_00, WEATHER_DAY_00),
                (INDEX1_22, ARRANGE_00, WEATHER_DAY_00)
        ]
        for try_index in try_indexes:
            #print 'Trying to find', resource_type, 'in', try_index,
            try:
                file_spec = self.other_files[try_index]
                #print file_spec,
                F = fft.map.resource.File(file_spec)
                #print F.toc[resource_type]
                if F.toc[resource_type] == 0: raise KeyError
                break
            except KeyError:
                file_spec = None
                F = None
                continue
        return F

    def get_texture(self, index):
        try_indexes = [
                index,
                (index[0], index[1], WEATHER_DAY_00),
                (index[0], ARRANGE_00, WEATHER_DAY_00),
                (INDEX1_30, ARRANGE_00, WEATHER_DAY_00),
                (INDEX1_22, ARRANGE_00, WEATHER_DAY_00)
        ]
        for try_index in try_indexes:
            try:
                file_spec = self.texture_files[try_index]
                break
            except KeyError:
                file_spec = None
        print 'Reading texture from', file_spec
        return file_spec

    def get_polygons(self, index, start):
        F = self.open_resource(index, start)
        if F == None: return []
        print 'Reading polygon XYZ coordinates from', F.name
        F.seek(F.toc[start])
        self.trianglesTextured = F.readShort(0)
        self.squaresTextured = F.readShort(0)
        self.trianglesUntextured = F.readShort(0)
        self.squaresUntextured = F.readShort(0)
        counts = {
                'textured': {3: self.trianglesTextured, 4: self.squaresTextured},
                'skirt': {3: self.trianglesUntextured, 4: self.squaresUntextured}
        }
        polygons = []
        for kind in ['textured', 'skirt']:
            for sides in [3, 4]:
                for zz in range(counts[kind][sides]):
                    triangle = F.readTriangle()
                    triangle['page'] = None
                    triangle['palette'] = None
                    triangle['vis'] = None
                    polygons.append(triangle)
                    if sides == 4:
                        # If reading squares, seek back to read the middle
                        # vertices again
                        F.seek(-12, 1)
                        triangle = self.wind_cw(F.readTriangle())
                        triangle['page'] = None
                        triangle['palette'] = None
                        triangle['vis'] = None
                        polygons.append(triangle)
        print 'Reading polygon normal vectors from', F.name
        i = 0
        for sides in [3, 4]:
            for zz in range(counts['textured'][sides]):
                light = F.readTriangleLight()
                polygons[i]['a'].update(light['a'])
                polygons[i]['b'].update(light['b'])
                polygons[i]['c'].update(light['c'])
                if sides == 4:
                    i += 1
                    # If reading squares, seek back to read the middle
                    # vertices again
                    F.seek(-12, 1)
                    light = self.wind_cw(F.readTriangleLight())
                    polygons[i]['a'].update(light['a'])
                    polygons[i]['b'].update(light['b'])
                    polygons[i]['c'].update(light['c'])
                i += 1
        i = 0
        print 'Reading polygon UV coordinates from', F.name
        for sides in [3, 4]:
            for zz in range(counts['textured'][sides]):
                uvs = F.readSquTexture()
                polygons[i]['a'].update(uvs['a'])
                polygons[i]['b'].update(uvs['b'])
                polygons[i]['c'].update(uvs['c'])
                polygons[i]['page'] = uvs['page'] & 0x03
                polygons[i]['palette'] = uvs['palette']
                if sides == 3:
                    # If reading triangles, then the readSquTexture call put us
                    # partway into the next triangle, so seek back a little
                    F.seek(-2, 1)
                else:
                    i += 1
                    polygons[i]['a'].update(uvs['d'])
                    polygons[i]['b'].update(uvs['c'])
                    polygons[i]['c'].update(uvs['b'])
                    polygons[i]['page'] = uvs['page'] & 0x03
                    polygons[i]['palette'] = uvs['palette']
                i += 1
        return polygons

    def get_polygons_anim(self, index):
        F = self.open_resource(index, fft.map.TOC_MESH_EXTRA_ANIM)
        if F == None: return []
        print 'Reading extra polygon placement instructions from', F.name
        F.seek(F.toc[fft.map.TOC_MESH_EXTRA_ANIM] + 8)
        for i in range(128):
            instruct = {}
            rot = [ang / 4096.0 for ang in struct.unpack('<hhhxx', F.read(8))]
            if rot[0]:
                instruct['angle'] = rot[0]
                instruct['axis'] = (1.0, 0.0, 0.0)
            elif rot[1]:
                instruct['angle'] = -rot[1]
                instruct['axis'] = (0.0, 1.0, 0.0)
            elif rot[2]:
                instruct['angle'] = rot[2]
                instruct['axis'] = (0.0, 0.0, 1.0)
            else:
                instruct['angle'] = None
                instruct['axis'] = None
            instruct['trans'] = struct.unpack('<hhhxx', F.read(8))
            instruct['trans'] = (instruct['trans'][0], -instruct['trans'][1], instruct['trans'][2])
            instruct['scale'] = [sf / 4096.0 for sf in struct.unpack('<hhhxx', F.read(8))]
            self.polygons_instructions.append(instruct)
            F.seek(56, 1)
        F.seek(8, 1)
        for i in range(len(self.polygons_extra)):
            self.polygons_extra_transform.append([])
            block_index = ord(F.read(1)) - 1
            self.polygons_extra_transform[i].append(self.polygons_instructions[block_index])
            F.seek(63, 1)
        F.seek(F.toc[fft.map.TOC_MESH_EXTRA_ANIM] + 0x3818)
        for i in range(len(self.polygons_extra)):
            parent_mesh = struct.unpack('<I', F.read(4))[0]
            if parent_mesh:
                for instruction in self.polygons_extra_transform[parent_mesh - 1]:
                    self.polygons_extra_transform[i].append(instruction)

    def wind_cw(self, tri):
        (tri['a'], tri['c']) = (tri['c'], tri['a'])
        return tri

    def get_polygon_visibility(self, index):
        F = self.open_resource(index, fft.map.TOC_MESH_VISIBILITY)
        if F == None: return
        print 'Reading polygon visibility info from', F.name
        counts = {
                'textured': {3: self.trianglesTextured, 4: self.squaresTextured},
                'skirt': {3: self.trianglesUntextured, 4: self.squaresUntextured}
        }
        block_beginnings = {
                'textured': {3: 0x000, 4: 0x400},
                'skirt': {3: 0xA00, 4: 0xA80}
        }
        for kind in ['textured', 'skirt']:
            for sides in [3, 4]:
                F.seek(F.toc[fft.map.TOC_MESH_VISIBILITY] + 0x380 + block_beginnings[kind][sides])
                for i in range(counts[kind][sides]):
                    vis = F.readShort()
                    yield [ (vis & 2**(15-x)) >> (15-x) for x in range(16) ]
                    if sides == 4:
                        yield [ (vis & 2**(15-x)) >> (15-x) for x in range(16) ]

    def get_lights(self, index):
        F = self.open_resource(index, fft.map.TOC_LIGHTS_BACKGROUND)
        if F == None: return
        print 'Reading lights from', F.name
        return F.getLights()

    def get_background(self, index):
        F = self.open_resource(index, fft.map.TOC_LIGHTS_BACKGROUND)
        if F == None: return
        print 'Reading background from', F.name
        return F.getBGGrad()

    def get_terrain(self, index):
        F = self.open_resource(index, fft.map.TOC_TERRAIN)
        if F == None: return
        print 'Reading terrain from', F.name
        return F.getTerrain()

    def get_palettes_color(self, index):
        F = self.open_resource(index, fft.map.TOC_PALETTES_COLOR)
        if F == None: return
        print 'Reading color palettes from', F.name
        return (F.getGLPalettes(), F.getGLPalettesAlpha())

    def find_gns_index(self, index_tuple):
        (arrangement, weather, day_night) = index_tuple
        weather_find = (day_night << 7) | (weather << 4)
        i = 0
        for index in self.all_indexes:
            if index[1] == arrangement and index[2] == weather_find:
                return i
            i += 1
        return 0
